'''Nicolas Trinephi - acse-nt719

Pandas DataFrame creator and manipulator

This module allows the user to load data into pd.DataFrames.

This module accepts space separated value files (.dat).

Contains pandas_data class
'''

from vfm_tool.Utils import *
import pandas as pd
import copy

class pandas_data:
  '''
  Class for data loading and manipulation with pandas
  
  Attributes:
    df: dataframe containing data
    list_df: list of dataframes
      contains dataframes per well name
    list_ids: list of str
      contains well names
      
  Methods:
    make_df()
      Creates pandas dataframes from historic data files
    df_lister(id_col)
      Makes list of pd.DataFrames, one for each unique entre of id_col
    make_stack(day, month, test)   static method
      Combines pd.DataFrames for plotting. The dataframe sizes and well names must match for plotting
    process(o_day, o_month, o_test, smooth=False)
      Processes the data using combination of filtering functions
    high_freq_local_only()
      Loads high frequency pressure data into dataframes using excel, usable on local machines but not on databricks
  '''
  
  def __init__(self, file_location):
    self.file_location = file_location
    self.df = None
    self.list_df = None
    self.list_ids = None
    
  def make_df(self):
    '''
    Creates pandas dataframes from historic data files
    
    Parameters:
      file_location: str
        directory of file
      datatype: str
        type of historic data
    Outputs:
      df: pd.DataFrame
        dataframe containing the loaded data
    '''

    if 'FileStore' in self.file_location: # check if local or on databricks
      file = '/dbfs' + self.file_location
    else:
      file = self.file_location

    if '_TEST_' in self.file_location:
      print('***********Accepted test data***********')
      utils.re_writer(self.file_location)
      self.df = pd.read_csv(file, header=0, sep=' ', engine='python', skiprows=2, index_col=False)
      self.df.drop(['Unnamed: 32', '*RESULT_NO'], axis=1, inplace=True)
      self.df['*DATE'] = pd.to_datetime(self.df['*DATE'])
      self.df = self.df.rename(columns={'*OIL_RATE': '*rOIL'})
      self.df = self.df.rename(columns={'*WAT_RATE': '*WATER'})
      self.df = self.df.rename(columns={'*WHP': '*WHP_AVG_P'})
      self.df = self.df.rename(columns={'*WHT': '*WHT_AVG_P'}).sort_values('*WELL_CODE')

    elif 'MONTH' in self.file_location:
      print('***********Accepted month data***********')
      utils.re_writer(self.file_location)
      self.df = pd.read_csv(file, header=0, sep=' ', engine='python', skiprows=2, index_col=False)
      self.df.drop(['Unnamed: 24'], axis=1, inplace=True)
      self.df['*DATE'] = pd.to_datetime(self.df['*DATE'])
      self.df = self.df.rename(columns={'*GAS': '*rGAS'})
      self.df = self.df.rename(columns={'*WHP_P': '*WHP_AVG_P'})
      self.df = self.df.rename(columns={'*WHT_P': '*WHT_AVG_P'}).sort_values('*WELL_BORE_CODE')

    elif '_DAY_' in self.file_location:
      print('***********Accepted day data***********')
      utils.re_writer(self.file_location)
      self.df = pd.read_csv(file, header=0, sep=' ', engine='python', skiprows=2, index_col=False)
      self.df.drop(['Unnamed: 28'], axis=1, inplace=True)
      self.df['*DATE'] = pd.to_datetime(self.df['*DATE'])
      self.df = self.df.rename(columns={'*GAS': '*rGAS'})
      self.df = self.df.rename(columns={'*OIL': '*rOIL'}).sort_values('*WELL_BORE_CODE')

    elif 'DAY.' in self.file_location:
      print('Choose a more recent file.')
      return
    else:
      try:
        raise InputError(self.file_location)
      except:
        print('InputError,', self.file_location, 'is not a good file.')
        
  def df_lister(self, id_col):
    '''
    Makes list of pd.DataFrames, one for each unique entre of id_col
    
    Paramters:
      df: pd.DataFrame
        original dataframe to be made into list
      id_col: str
        column with entries to be listed
      date: str
        date column
    Output:
      df_list: list of pd.DataFrames
        list containing data frames of each id_col
      list_ids: list of str
        list containing unique entries of id_col     
    '''
    dict_ids = self.df[id_col].unique().tolist()
    self.list_ids = []
    for i in range(len(dict_ids)):
      self.list_ids.append(str(dict_ids[i])) # keep str values
    # create list of DataFrames
    self.list_df = [self.df.where(self.df[id_col] == x) for x in self.list_ids]
    for i in range(len(self.list_df)):
      self.list_df[i] = self.list_df[i].dropna(subset=[id_col])
      self.list_df[i] = self.list_df[i].sort_values(['*DATE'])

  def make_stack(o_day, o_month, o_test):
    '''
    Combines pd.DataFrames for plotting. The dataframe sizes and well names must match for plotting.
    
    day, month and test objects must be created
    
    Parameters:
      day: data object
        list of day dataframes    
      month: data object
        list of month dataframes    
      test: data object
        list of test dataframes
    Outputs:
      day: pd.DataFrame
        day dataframes together by tail name
      month: pd.DataFrame
        month dataframes together by tail name
      test: pd.DataFrame
        test dataframes together by tail name
    '''

    day = o_day # shallow copy for spark compatibility
    month = o_month
    test = o_test
    # change to pd.DataFrame if spark_data uses this method
    if not isinstance(o_day.df, pd.DataFrame):
      if not isinstance(o_day.list_df[0], pd.DataFrame):
        day.list_df = utils.to_pandas(o_day.list_df)
    if not isinstance(o_month.df, pd.DataFrame):
      if not isinstance(o_month.list_df[0], pd.DataFrame):
        month.list_df = utils.to_pandas(o_month.list_df)
    if not isinstance(o_test.df, pd.DataFrame):
      if not isinstance(o_test.list_df[0], pd.DataFrame):
        test.list_df = utils.to_pandas(o_test.list_df)


    # restack data and prepare for plotting
    index = []
    day_bore = []
    month_bore = []
    test_copies = []
    cnt = 0
    for n in range(len(test.list_ids)):
      index.append([index for index, name in enumerate(day.list_ids) if test.list_ids[n] in name]) # get index of test_wells
      # stack dataframes for each well
      day_bore.append([day.list_df[index[n][i]] for i in range(len(index[n]))])
      month_bore.append([month.list_df[index[n][i]] for i in range(len(index[n]))])
      day_bore[n] = pd.concat(day_bore[n], ignore_index=True)
      month_bore[n] = pd.concat(month_bore[n], ignore_index=True)
      # expand test data for plotting, add well bore names 
      tmp = [] # refresh temp to match day and month bore names
      tmp_df = test.list_df[n].copy()
      tmp_df['*WELL_BORE_CODE'] = day.list_ids[cnt]
      tmp.append(tmp_df)
      cnt += 1
      for k in range(len(index[n])-1):
        tmp_df = test.list_df[n].copy()
        tmp_df['*WELL_BORE_CODE'] = day.list_ids[cnt]
        tmp.append(tmp_df)
        cnt += 1
      test_copies.append(tmp)
      test_copies[n] = pd.concat(test_copies[n], ignore_index=True)
    # stack wells together for plotting
    day_ = pd.concat(day_bore, ignore_index=True)
    month_ = pd.concat(month_bore, ignore_index=True)
    test_ = pd.concat(test_copies, ignore_index=True)
    return day_, month_, test_

  def process(o_day, o_month, o_test, smooth=False):
    '''
    Processes the data using combination of filtering functions
    
    Previous functions:clear data outside of limits, make list of dataframes, 
    calculate unaccumulated monthly rates, apply filter if needed, concat dataframes
    together for plotting.
    
    Parameters:
      o_day: pandas_data object
        object containing original day data
      o_month: pandas_data object
        object containing original month data
      o_test: pandas_data object
        object containing original test data
    Outputs:
      day: pandas_data object
        object containing processed day data
      month: pandas_data object
        object containing processed month data
      test: pandas_data object
        object containing processed test data
    '''
    # deep copy of the original objects
    if isinstance(o_day.df, pd.DataFrame) and isinstance(o_day.list_df[0], pd.DataFrame): # makes copy
      day = copy.deepcopy(o_day)
    if isinstance(o_month.df, pd.DataFrame) and isinstance(o_month.list_df[0], pd.DataFrame):
      month = copy.deepcopy(o_month)
    if isinstance(o_test.df, pd.DataFrame) and isinstance(o_test.list_df[0], pd.DataFrame):
      test = copy.deepcopy(o_test)

    # change to pd.DataFrame if spark_data uses this method, deep copy not available
    if not isinstance(o_day.df, pd.DataFrame):
      day = o_day
      day.df  = o_day.df.toPandas()
      if not isinstance(o_day.list_df[0], pd.DataFrame):
        day.list_df = utils.to_pandas(o_day.list_df)
    if not isinstance(o_month.df, pd.DataFrame):
      month = o_month
      month.df  = o_month.df.toPandas()
      if not isinstance(o_month.list_df[0], pd.DataFrame):
        month.list_df = utils.to_pandas(o_month.list_df)
    if not isinstance(o_test.df, pd.DataFrame):
      test = o_test
      test.df = o_test.df.toPandas()
      if not isinstance(o_test.list_df[0], pd.DataFrame):
        test.list_df = utils.to_pandas(o_test.list_df)

    utils.nan_exc(day.df, month.df, test.df)

    for i in range(len(test.list_df)):
      test.list_df[i] = test.list_df[i].sort_values('*DATE')

    for i, df in enumerate(month.list_df):
      month.list_df[i] = utils.un_accum(df)

    for i in range(len(day.list_df)):
      day.list_df[i] = day.list_df[i].sort_values('*DATE').reset_index(drop=True)

      if smooth:
        month.list_df[i] = utils.dynamic_filter(month.list_df[i], 30, '*BHP_AVG', 10)
        month.list_df[i] = utils.dynamic_filter(month.list_df[i], 10, '*BHT_AVG', 50)
        month.list_df[i] = utils.dynamic_filter(month.list_df[i], 10, '*WHP_AVG_P', 50)
        month.list_df[i] = utils.dynamic_filter(month.list_df[i], 10, '*WHT_AVG_P', 50)
        month.list_df[i] = utils.dynamic_filter(month.list_df[i], 10, '*rOIL', 50)
        month.list_df[i] = utils.dynamic_filter(month.list_df[i], 10, '*rGAS', 50)
        month.list_df[i] = utils.dynamic_filter(month.list_df[i], 10, '*WATER', 50)

        day.list_df[i] = utils.dynamic_filter(day.list_df[i], 30, '*BHP_AVG', 10)
        day.list_df[i] = utils.dynamic_filter(day.list_df[i], 10, '*BHT_AVG', 50)
        day.list_df[i] = utils.dynamic_filter(day.list_df[i], 10, '*WHP_AVG_P', 50)
        day.list_df[i] = utils.dynamic_filter(day.list_df[i], 10, '*WHT_AVG_P', 50)
        day.list_df[i] = utils.dynamic_filter(day.list_df[i], 10, '*rOIL', 50)
        day.list_df[i] = utils.dynamic_filter(day.list_df[i], 10, '*rGAS', 50)
        day.list_df[i] = utils.dynamic_filter(day.list_df[i], 10, '*WATER', 50)
    day.df, month.df, test.df = pandas_data.make_stack(day, month, test)
    return day, month, test
  
  def high_freq(df_high, C2_bore):
    '''
    Loads high frequency pressure data into dataframes using excel, usable on local machines but not on databricks.
    
    Outputs:
      high_freq_p: pd.DataFrame
        contains pressure values with other well names empty for plotting
      high_freq_oil: pd.DataFrame
        contains oil rate values with other well names empty for plotting
      df_high: 
        contains all high frequency data
    '''

    # trim excel data 
    df_high[0] = df_high[0][['Date', 'Pressure']]
    df_high[1] = df_high[1][['Date', 'Pressure']]
    df_high[2] = df_high[2][['Date', 'Pressure']]
    df_high[3] = df_high[3][['Date', 'OIL']]
    df_high[4] = df_high[4][['Time @ end', 'Cumulative Volume']]
    df_high[5] = df_high[5][['Time @ end', 'Cumulative Volume']]

    df_high[4] = df_high[4].rename(columns={'Time @ end': '*DATE'})
    df_high[4]['*WELL_BORE_CODE'] = C2_bore[0]
    df_high[5] = df_high[5].rename(columns={'Time @ end': '*DATE'})
    df_high[5]['*WELL_BORE_CODE'] = C2_bore[1]

    for i in range(3):
        df_high[i] = df_high[i].rename(columns={'Date': '*DATE', 'Pressure': '*DHP'})
        df_high[i]['*WELL_BORE_CODE'] = C2_bore[i]
    df_high[3] = df_high[3].rename(columns={'Date': '*DATE'})

    # remove unit cells
    for i in range(6):
        df_high[i].drop(df_high[i].tail(1).index, inplace=True)
        df_high[i].drop(df_high[i].head(1).index, inplace=True)
        df_high[i] = df_high[i].sort_values('*DATE', ascending=True)

    df_high[4].loc[1, '*rOIL'] = df_high[4].loc[1, 'Cumulative Volume']
    df_high[5].loc[1, '*rOIL'] = df_high[5].loc[1, 'Cumulative Volume']

    for i in range(2, len(df_high[4])):
        df_high[4].loc[i, '*rOIL'] = df_high[4].loc[i, 'Cumulative Volume'] - df_high[4].loc[i-1, 'Cumulative Volume']
        df_high[5].loc[i, '*rOIL'] = df_high[5].loc[i, 'Cumulative Volume'] - df_high[5].loc[i-1, 'Cumulative Volume']

    # merge df_high together
    high_freq_p = pd.concat([df_high[0], df_high[1], df_high[2]], ignore_index=True)
    high_freq_oil = pd.concat([df_high[4], df_high[5]], ignore_index=True)

    # set up dataframe to plot with other data
    file_day = '../Data/General/HYME_DAILY_PRODUCTION/NJORD_OFM_HIST3_DAY_20131113_143735.dat' # use your directory
    df_day = make_df(file_day, 'daily').sort_values('*WELL_BORE_CODE')
    day_list, well_day_ids = df_lister(df_day, '*WELL_BORE_CODE', '*DATE')
    t = np.array(well_day_ids)
    tmp = pd.DataFrame()
    tmp['*WELL_BORE_CODE'] = t
    high_freq_p = pd.concat([high_freq_p, tmp], axis=0).sort_values('*WELL_BORE_CODE')
    high_freq_oil = pd.concat([high_freq_oil, tmp], axis=0).sort_values('*WELL_BORE_CODE')
    return high_freq_p, high_freq_oil, df_high



# COMMAND ----------





'''Nicolas Trinephi - acse-nt719

Module with quality control functions

Functions:
  c_desc
    Adds and removes rows to make a custom description dataframe
  qc_list(input, ids)
    creates dataframe of concatenated custom describe dataframes from a list of dataframes
  qc_data
    Provides statistical analysis of data in file_location

'''

from vfm_tool.Utils import *
from vfm_tool.Pandas_data import *
from vfm_tool.Spark_data import *
import copy
import pandas as pd

def c_desc(input, desc):
  '''
  Creates a custom description dataframe

  Parameters:
    pandf: pd.DataFrame
      input dataframe
    df_l: pd.DataFrame
      description dataframe of pandf
  Output:
    df_l: pd.DataFrame
      custom description dataframe
  '''
  
  # make new rows
  df = desc.copy()
  df.loc['range'] = df.loc['min'].round(2).apply(str) + ' - ' + df.loc['max'].round(2).apply(str)
  df.loc['% null'] = (input == 0).astype(int).sum(axis=0) * 100 / len(input)
  # keep useful rows
  df = df.loc[['count', 'mean', 'range', '% null']].round(2)
  return df

def qc_list(input, ids):
  '''
  creates dataframe of concatenated custom describe dataframes from a list of dataframes.

  Parameter:
    df_list: list of pd.DataFrames
  Output:
    df: pd.DataFrames
  '''

  df = []
  for i in range(len(input)):
    df.append(input[i].describe())
    df[i] = c_desc(input[i], df[i])
  df = pd.concat(df, keys=ids, sort=True)
  return df

def qc_data(obj):
  '''
  Provides statistical analysis of data in file_location. Accepts daily, monthly and test data with units.
  
  Parameters:
    obj: data object, spark_data() or pandas_data()
      location of data file.

  Outputs:
    df: pd.DataFrame
      Contains statistical elements necessary for quality control of the data.
  '''

  # check object type and create pd.DataFrame if spark_data is used.
  if isinstance(obj, spark_data):
    tmp_list = utils.to_pandas(obj.list_df)
    df = qc_list(tmp_list, obj.list_ids)
  elif isinstance(obj, pandas_data):
    df = qc_list(obj.list_df, obj.list_ids)
  else:
    raise TypeError('The input is not a valid type, please input a spark_data() or pandas_data() object.')

  # add units
  if '_DAY.' in obj.file_location:
    print('It would be better to choose a more recent file.')
    return
  elif '_DAY_' in obj.file_location:
    print('*********** qc day data ***********')
    # add units heading
    df.loc[('units', ''),:] = ('hours', '   bar', '\N{DEGREE SIGN}C', 'bar', 'bar', '%', 'm\u00b3/day', 'bar', '\N{DEGREE SIGN}C', '$ US', '%', 'm\u00b3/day', 'm\u00b3/day', \
                                     'm\u00b3/day', 'm\u00b3/day', 'm\u00b3/day', 'm\u00b3/day', 'bar', '\N{DEGREE SIGN}C', '%', '%', 'bar', 'bar', '\N{DEGREE SIGN}C', '\N{DEGREE SIGN}C')

  elif 'MONTH' in obj.file_location: # month data
    print('*********** qc month data ***********')
    # add units heading
    df.loc[('units', ''),:] = ('m\u00b3/day', 'm\u00b3/day', 'm\u00b3/day', 'm\u00b3/day', 'hours', 'bar', '\N{DEGREE SIGN}C', 'bar', 'bar', 'm\u00b3', '$ US', '%', \
                                     'm\u00b3/day', '%', '%', '%', 'bar', 'bar', 'bar', '\N{DEGREE SIGN}C', '\N{DEGREE SIGN}C', '\N{DEGREE SIGN}C')

  elif '_TEST_' in obj.file_location: # well test data
    print('*********** qc test data ***********')
    # add units heading
    df.loc[('units', ''),:] = ('%', 'm\u00b3/day', 'bar', 'm\u00b3/day', 'm\u00b3/day', 'm\u00b3/day', '$ US', '$ US', 'bar', \
                                     '\N{DEGREE SIGN}C', '$ US', 'm\u00b3/day', 'm\u00b3/day', 'm\u00b3/day', 'metric', 'm\u00b3', 'm\u00b3', 'm\u00b3', 'm\u00b3',\
                                     'metric', '$ US', 'bar', '\N{DEGREE SIGN}C', '$ US', 'bar', '\N{DEGREE SIGN}C', 'm\u00b3/day')
  else:
    raise TypeError('Not a valid data file.')
  df = df.T.set_index('units', append=1).T
  return df

def qc_high_freq(high_freq, sheet_names):
  '''
  Obtains statistical analysis of the high frequency data.

  Parameters:
    high_freq: list pd.DataFrame
      Dataframe containing high frequency data  
    sheet_names: list str
      well tail names
    Outputs:
      desc_well: pd.DataFrame
        Dataframe containing custom description of pressures at each tail
      desc_oil: pd.DataFrame
        Dataframe containing custom description of oil rate at well head.
  '''

  df = copy.deepcopy(high_freq) # deep copy of the list of df

  # for pressures
  df_l = []
  for i in range(len(df)-1):
    df[i].drop(df[i].columns[2], axis=1, inplace=True)
    df[i].columns = ['*DATE', 'Pressure']
    df_l.append(df[i].describe())
    df_l[i] = c_desc(df[i], df_l[i])
  desc_well = pd.concat(df_l, keys=sheet_names)
  # add units heading
  desc_well.loc[('units', ''),:] = ('bar')
  desc_well = desc_well.T.set_index('units', append=1).T

  # for oil rate
  desc_oil = df[3].describe()
  desc_oil = c_desc(df[3], desc_oil)
  desc_oil.loc[('units')] = ('m\u00b3/day')
  desc_oil = desc_oil.T.set_index('units', append=1).T
  return desc_well, desc_oil



'''Nicolas Trinephi - acse-nt719

Module Containing visualization functions.

Requires plotly, hvplot, holoviews and panel installed. Requires notebook widgets for holoviews
visualizations. Jupyter Notebook is known to work best.

Functions:
  data_plot(df_day, df_month, df_test, x_val ,y_val, title, day='Day', month='Month', test='Test')
    Plots day, month and test data together on the same axis
  holo_plot(df_day, df_month, df_test)
    Interactive plotting of the historic data, parameters against date
  hf_plot(df_high)
    high frequency pressure and oil rate plot using plotly
  scatter_plot(df_day, df_month, df_test)
    Plots parameters against parameter and oil rate against parameter to check for outliers.
  scatter_plot
    Creates scatter plots of pressure vs temperature and oil rate against pressures and temperatures
  
'''

import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import plotly.offline as pyo
import panel as pn
import hvplot
import hvplot.pandas
import holoviews as hv
import panel as pn

# parameter names
oil_rate = '*rOIL'
gas_rate = '*rGAS'
water_rate = '*WATER'
BHP = '*BHP_AVG'
BHT = '*BHT_AVG'
WHP = '*WHP_AVG_P'
WHT = '*WHT_AVG_P'
code = '*WELL_BORE_CODE'
date = '*DATE'

#global plotting functions
def data_plot(df_day, df_month, df_test, x_val ,y_val, title, day='Day', month='Month', test='Test'):
  '''
  Plots day, month and test data together on the same axis

  Parameters:
    df_day: pd.DataFrame
      dataframes containing day data
    df_month: pd.DataFrame
      dataframes containing month data
    df_test: pd.DataFrame
      dataframes containing test data
    x_val: str
      column to be used as x values
    y_val: str
      column to be used as y values
    title: str
      title of plot
    day: str default: 'Day'
      label of day line
    month: str default: 'Month'
      label of month line
    test: str default: 'Test'
      label of test line

  Output:
    plot: hvplot
      interactive plot with day, month and test lines on the same axis
  '''

  day_plot = df_day.hvplot.scatter(x=x_val, y=y_val, groupby=code, height=600, width=1000, label=day, title=title)
  month_plot = df_month.hvplot.scatter(x=x_val, y=y_val, groupby=code, label=month)
  test_plot = df_test.hvplot.scatter(x=x_val, y=y_val, groupby=code, label=test)
  plot = day_plot * month_plot * test_plot
  return plot

def holo_plot(df_day, df_month, df_test):
  '''
  Interactive plotting of the historic data, parameters against date

  Parameters:
    df_day: pd.DataFrame
      dataframes containing day data
    df_month: pd.DataFrame
      dataframes containing month data
    df_test: pd.DataFrame
      dataframes containing test data

  Output:
    plot: hvpot
      interactive plot of historic data, parameters against date in two columns  
  '''

  # oil graph
  oil = data_plot(df_day, df_month, df_test, date, oil_rate, 'OIL').opts(ylabel='Oil Rate (m\u00b3/day)')

  # gas graph
  gas = data_plot(df_day, df_month, df_test, date, gas_rate, 'GAS').opts(ylabel='Gas Rate (m\u00b3/day)')

  # water graph
  water = data_plot(df_day, df_month, df_test, date, water_rate, 'WATER').opts(ylabel='Water Rate (m\u00b3/day)')

  # pressure
  back_pressure = data_plot(df_day, df_month, df_test, date, BHP, 'Pressure', day='Day BHP', month='Month BHP', test='Test BHP')
  work_pressure = data_plot(df_day, df_month, df_test, date, WHP, 'Pressure', day='Day WHP', month='Month WHP', test='Test WHP')
  pressure = back_pressure * work_pressure
  pressure = pressure.opts(ylabel='Pressure (bar)')

  # temperature
  back_temp = data_plot(df_day, df_month, df_test, date, BHT, 'Temperature', day='Day BHT', month='Month BHT', test='Test BHT')
  work_temp = data_plot(df_day, df_month, df_test, date, WHT, 'Temperature', day='Day WHT', month='Month WHT', test='Test WHT')
  temp = back_temp * work_temp
  temp = temp.opts(ylabel='Temperature (\N{DEGREE SIGN}C)')
  plot = (oil + gas + water + pressure + temp).cols(2)
  return plot
  
def hf_plot(df_high, bore):
  '''
  high frequency pressure and oil rate plot using plotly

  Parameters:
    df_high: list of pd.DataFrames
      list of high frequency data frames, [3] is expected to be oil rate
    bore: list of str
      bore names
  ''' 

  fig = make_subplots(specs=[[{"secondary_y": True}]])
  fig.update_layout(
      autosize=True,
      height=800)
  for i in range(3):
    fig.add_trace(go.Scattergl(x=df_high[i]['*DATE'], y=df_high[i]['*DHP '+bore[i]], name=bore[i], mode='markers'))
  fig.add_trace(go.Scattergl(x=df_high[3]['*DATE'], y=df_high[3]['*rOIL'], name='Oil Rate', mode='lines+markers',\
                               marker_color='black'), secondary_y=True)
  
  fig.update_layout(
    title="High Frequency Data",
    xaxis_title="Date",
    yaxis_title='Pressure (bar)',
    yaxis2_title="Rate (m\u00b3)",
    font=dict(
      family="Courier New, monospace",
      size=26,
      color="#7f7f7f"))
  fig.show()
  
def scatter_plot(df_day, df_month, df_test):
  '''
  Plots parameters against parameter and oil rate against parameter to check for outliers.

  Parameters:
    df_day: pd.DataFrame
      dataframe of day data
    df_month: pd.DataFrame
      dataframe of month data
    df_test: pd.DataFrame
      dataframe of test data

  Output:
    plot: hvplot
      scatter plots in two columns
  '''

  wpt = data_plot(df_day, df_month, df_test, WHT, WHP, 'Pressure-Temperature', day='Day BHP', month='Month BHP', test='Test BHP')
  bpt = data_plot(df_day, df_month, df_test, BHT, BHP, 'Pressure-Temperature', day='Day BHP', month='Month BHP', test='Test BHP')
  owp = data_plot(df_day, df_month, df_test, oil_rate, WHP, 'oil WHP', day='Day WHP', month='Month WHP', test='Test WHP')
  obp = data_plot(df_day, df_month, df_test, oil_rate, BHP, 'oil BHP', day='Day BHP', month='Month BHP', test='Test BHP')
  owt = data_plot(df_day, df_month, df_test, oil_rate, WHT, 'oil WHT', day='Day WHT', month='Month WHT', test='Test WHT')
  obt = data_plot(df_day, df_month, df_test, oil_rate, BHT, 'oil BHT', day='Day BHT', month='Month BHT', test='Test BHT')
  plot = (wpt + bpt + owp + obp + owt + obt).cols(2)
  return plot

# COMMAND ----------





'''Nicolas Trinephi - acse-nt719

Keras LSTM trainer

This module allows the user to input a dataset, prepare it for insertion into keras LSTM algorithm, train an LSTM network 
and predict from a test set.

Inputs should be pd.DataFrames.

This script requires all libraries from requirements.txt to be installed on the cluster/machine.

Contains VFM_LSTM class.
'''

from vfm_tool.Utils import *

# import machine learning libraries
from keras.models import Sequential
from keras.layers import Dense
from keras.layers import LSTM
from sklearn.preprocessing import LabelEncoder, MinMaxScaler, StandardScaler
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score
import tensorflow as tf
from keras.optimizers import Adam
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Bidirectional, Dense, LSTM, Dropout
from tensorflow.python.keras.layers.advanced_activations import LeakyReLU
from tensorflow.keras.callbacks import Callback, ModelCheckpoint, EarlyStopping
from matplotlib import pyplot
import numpy as np
import mlflow
import mlflow.keras
  
class VFM_LSTM():
  '''
  Normalizes and trains an LSTM model to predict a test set.
  
  Attributes:
    RANDOM_SEED: int
      set random seed for wieght initialization
    input: pd.DataFrame
      dataframe of the problem
    name: str
      name of model
    train_size: int or float (usually int)
      percentage of the samples to be used as training set
    train: pd.DataFrame
      train set
    test: pd.DataFrame
      test set
    num_labels: int
      number of labels in the input
    list_labels: list of str
      label names
    f_transformer: sklearn.preprocessing.StandardScaler()
      normalization of training features
    l_transformer: sklearn.preprocessing.StandardScaler()
      normalization of training labels
    f_cols: int
      number of feature columns
    l_cols: int
      number of label columns
    X_train: np.array
      training set features
    X_test: np.array
      test set features
    y_train: np.array
      training set labels
    y_test: np.array
      test set labels
    model: Sequential()
      keras model
    history: model.fit()
      training the model
    y_pred_inv: np.array
      unscaled prediction array
    rates: str
      name of labels
    metrics: float
      calculated metrics of the prediction
      
  Methods
    split_data(split_prcnt)
      splits the input data into training and test set
    normalize()
      rescales the data with sklearn.preprocessing.StandScaler()
    to_supervised(lbck)
      wrapper for c_dataset, adds lookback
    c_dataset(X, y lbck=1)  static method
      creates supervised learning data set with lookback 
    train_model(units, epochs, batch_size, lr, drop, mode)
      trains keras LSTM model
    loss_model()
      plots the training loss and validation loss
    predict_y()
      predict the test set and calculate metrics
    eval_metrics(actual, pred)   static method
      calculate rmse, mae and r2 score
    predict_plot(combined=True)
      plot prediction
    log_mlflow(log_dir)
      save model parameters and metrics in log_dir
  '''
  
  RANDOM_SEED = 42
  np.random.seed(RANDOM_SEED)
  tf.random.set_seed(RANDOM_SEED)
  tf.keras.backend.set_floatx('float64')

  def __init__(self, input, name):
    '''
    Parameters:
      input: pd.Dataframe
        dataframe of the problem
      name: str
        name of model
    '''
    
    self.input = input
    self.dataset = input.copy() # we copy the data so that the original data is not changed
    self.name = name
    self.train_size = None
    self.train = None
    self.test = None
    self.num_labels = None
    self.list_labels = None
    self.f_transformer = None
    self.l_transformer = None
    self.f_cols = None
    self.l_cols = None
    self.X_train = None
    self.X_test = None
    self.y_train = None
    self.y_test = None
    self.model = None
    self.history = None
    self.y_pred_inv = None
    self.rates = None
    self.metrics = None
    self.mode = None
    self.lr = None
    self.units = None
    self.num_epoch = None
    self.batch_size = None
    self.drop = None
    
    print('Data loaded and copied.')
    
  def split_data(self, split_prcnt):
    '''
    splits the input data into training and test set
    
    Parameter:
      split_prcnt: int or float (usually int)
        percentage of data to make training set
    '''

    if split_prcnt >= 100:
      raise Exception('Pick a number between 0 and 100 exclusive')
    if split_prcnt <= 0:
      raise Exception('Pick a number between 0 and 100 exclusive')

    train_split = split_prcnt/100 
    self.train_size = int(len(self.dataset) * train_split)
    test_size = len(self.dataset) - self.train_size
    self.train, self.test = self.dataset.iloc[0:self.train_size], self.dataset.iloc[self.train_size:len(self.dataset)] 
    print('Data split:')
    print(' ', split_prcnt, '% training : ', len(self.train))
    print(' ', 100 - split_prcnt, '% test : ', len(self.test))
    print('\n')
    
  def normalize(self):
    '''
    Rescales data and finds labels
    '''
    # check how many labels we have
    self.num_labels = 0
    self.list_labels = []
    
    if '*rOIL' in self.input.columns:
      self.num_labels += 1
      self.list_labels.append('*rOIL')
    if '*rGAS' in self.input.columns:
      self.num_labels += 1
      self.list_labels.append('*rGAS')
    if '*WATER' in self.input.columns:
      self.num_labels += 1
      self.list_labels.append('*WATER')
    if self.num_labels == 0:
      raise TypeError('There are no labels in the input dataset.')
      
    self.rate = []
    for label in self.list_labels:
      # replace labels for nice plotting
      if label == '*rGAS':
        self.rate.append('Gas Rate')
      elif label == '*rOIL':
        self.rate.append('Oil Rate')
      elif label == '*WATER':
        self.rate.append('Water Rate')
      else:
        try:
          raise InputError(label)
        except:
          print('InputError: ', label, ' is not a label from the dataset.')
          
    cols = self.dataset.columns.tolist()
    
    # feature and labels
    self.f_cols = cols[:-self.num_labels] # right of dataframe is features
    self.l_cols = cols[-self.num_labels:] # left of dataframe is labels
    
    # normalization
    self.f_transformer = StandardScaler()
    self.l_transformer = StandardScaler()

    # apply scalers to features and labels l
    self.f_transformer = self.f_transformer.fit(self.train[self.f_cols].to_numpy())
    self.l_transformer = self.l_transformer.fit(self.train[self.l_cols].to_numpy())

    self.train.loc[:, self.f_cols] = self.f_transformer.transform(self.train[self.f_cols].to_numpy())
    self.train[self.l_cols] = self.l_transformer.transform(self.train[self.l_cols].to_numpy())
    self.test.loc[:, self.f_cols] = self.f_transformer.transform(self.test[self.f_cols].to_numpy())
    self.test[self.l_cols] = self.l_transformer.transform(self.test[self.l_cols].to_numpy())    
    
  def to_supervised(self, lbck):
    '''
    Creates supervised data sets
    Parameter:
      lbck: int
        amount of past data to add as features
    '''
    
    self.lbck = lbck
    self.X_train, self.y_train = self.c_dataset(self.train, self.train[self.l_cols], self.lbck)
    self.X_test, self.y_test = self.c_dataset(self.test, self.test[self.l_cols], self.lbck)
    
  
  @staticmethod
  def c_dataset(X, y, lbck=1): # writen by Alexis
    '''
    Reshapes datasets to sequences with lookback for supervised learning
    
    Parameters:
      X: pd.DataFrame
        Training dataset
      y: pd.DataFrame
        Test dataset
      lbck: int
        amount of past data to add as features
      
    Outputs:
      Xs: np.array
        features
      ys: np.array
        labels
    '''

    Xs, ys = [], []
    for i in range(len(X) - lbck):
        v = X.iloc[i:(i + lbck)].values
        Xs.append(v)        
        ys.append(y.iloc[i + lbck])
    return np.array(Xs), np.array(ys)

  def train_model(self, units, epochs, batch_size, lr, drop, mode):
    '''
    Trains LSTM model
    
    Parameters:
      units: int
        number of LSTM neurons
      epochs: int
        number of iterations
      batch_size: int
        section of data to input
      lr: float
        learning rate
      drop: float
        drop rate
      mode: str
        LSTM model to use
    ''' 

    self.mode = mode
    self.lr = lr
    self.units = units
    self.num_epoch = epochs
    self.batch_size = batch_size
    self.drop = drop

    self.model = Sequential()
    
    if self.mode == 'classic':
      self.model.add(LSTM(units, input_shape=(self.X_train.shape[1], self.X_train.shape[2])))
      print(self.mode, 'LSTM')
    elif self.mode == 'bidirectional':
      self.model.add(Bidirectional(LSTM(units),input_shape=(self.X_train.shape[1], self.X_train.shape[2])))
      print(mode, 'LSTM')      
    elif self.mode == 'stacked': # two layers, any more is not useful for the extended training time required
      self.model.add(LSTM(units, input_shape=(self.X_train.shape[1], self.X_train.shape[2]), return_sequences=True))
      self.model.add(LSTM(units, input_shape=(self.X_train.shape[1], self.X_train.shape[2])))          
      print(self.mode, 'LSTM')
    elif self.mode == 'stacked-bi':
      self.model.add(Bidirectional(LSTM(units, return_sequences=True),input_shape=(self.X_train.shape[1], self.X_train.shape[2])))
      self.model.add(Bidirectional(LSTM(units),input_shape=(self.X_train.shape[1], self.X_train.shape[2])))
      print(self.mode, 'LSTM')
    elif self.mode == 'bi+classic':
      self.model.add(Bidirectional(LSTM(units, return_sequences=True),input_shape=(self.X_train.shape[1], self.X_train.shape[2])))
      self.model.add(LSTM(units, input_shape=(self.X_train.shape[1], self.X_train.shape[2])))
      print(self.mode, 'LSTM')
    else: 
      try:
        raise InputError(self.mode)
      except:
        print('InputError: ', self.mode, ' is not a valid LSTM mode.')
    
    self.model.add(Dropout(self.drop))
    self.model.add(Dense(self.num_labels)) 
    
    adam = Adam(lr=lr)
    # Stop training when a monitored quantity has stopped improving.
    earlystop = [EarlyStopping(monitor="loss", min_delta = 0.00001, patience = 35, mode = 'auto', restore_best_weights=True)] 
    checkpointer = ModelCheckpoint(filepath=self.name + "_weights.hdf5", verbose=1, save_best_only=True, monitor='val_mean_squared_error')

    # Using regression loss function 'Mean Standard Error' and validation metric 'Mean Absolute Error'
    self.model.compile(loss='mse', optimizer='adam', metrics=['mae','MeanSquaredError'])
    
    # fit the model
    self.history = self.model.fit(self.X_train, self.y_train, \
                            epochs=self.num_epoch, \
                            validation_split=0.1, \
                            callbacks = [checkpointer, earlystop], \
                            verbose=2, \
                            shuffle=False, \
                            initial_epoch=0, \
                            batch_size=self.batch_size)
    
  def loss_plot(self): # written by Alexis
    '''
    Plot the training and validation loss.
    '''
    
    # Calculate the train loss and train metric, in this case mean absolute error
    train_loss = np.mean(self.history.history['loss'])
    train_mae = np.mean(self.history.history['mae'])

    #title = 'Train Loss: {0:.3f} Test Loss: {1:.3f}\n  Train MAE: {2:.3f}, Val MAE: {3:.3f}'.format(train_loss, score[0], train_mae, score[1])
    title = 'Model Loss'

    # Plot loss function
    fig = pyplot.figure()
    pyplot.style.use('seaborn')
    
    pyplot.rcParams.update({'font.size': 30})
    pyplot.plot(self.history.history['loss'], 'c-', label='train')
    pyplot.plot(self.history.history['val_loss'], 'm:', label='validation')
    pyplot.xlabel('Epochs')
    pyplot.ylabel('Loss')
    pyplot.title(title)
    pyplot.legend()
    pyplot.grid(True)
    fig.set_size_inches(w=7,h=7)
    pyplot.show()

  def predict_y(self):
    '''
    Predict the unseen test set and calculate metrics.
    
    Output:
      self.metrics: instance dictionary
        dictionary containing metrtics of the model's prediction
    '''
    
    self.model.load_weights(self.name + '_weights.hdf5')
    y_pred = self.model.predict(self.X_test)
    #Performing Inverse Scaling, to revert the normalization scale
    
    y_train_inv = self.l_transformer.inverse_transform(self.y_train)
    y_test_inv = self.l_transformer.inverse_transform(self.y_test)
    self.y_pred_inv = self.l_transformer.inverse_transform(y_pred)
    
    self.metrics = {} # save metrics in a nested dictionary.
    for i in range(len(self.list_labels)):
      rmse, mae, r2 = self.eval_metrics(y_test_inv[:,i].flatten(), self.y_pred_inv[:,i].flatten())
      self.metrics[self.rate[i]] = {'rmse' : rmse, 
                                     'mae' : mae, 
                                      'r2' : r2}
    return self.metrics
  
  @staticmethod
  def eval_metrics(actual, pred):
    '''
    Calculate rmse, mae and r2 score.
    
    Parameters:
      actual: np.array
        actual test set
      pred: np.array
        prediction of test set
    '''
    
    rmse = np.sqrt(mean_squared_error(actual, pred))
    mae = mean_absolute_error(actual, pred)
    r2 = r2_score(actual, pred)
    return rmse, mae, r2

  def predict_plot(self, DATE, combined=True):
    '''
    Plot the prediction.
    
    Parameters:
      DATE: pd.DataFrame
        date indexes
      combined: bool default True
        If true combines the lines of the predicted rates.
    '''
    
    data_H = self.input[self.l_cols]
    train_data_H = data_H[:self.train_size]
    valid_data_H = data_H[self.train_size:]
    # re-attach the dates
    # df train
    df_train = pd.DataFrame(train_data_H)
    df_train = pd.concat([DATE, df_train], axis=1, join='inner')

    # df actual
    df_actual = pd.DataFrame(valid_data_H)
    df_actual = pd.concat([DATE, df_actual], axis=1, join='inner')

    # df prediction, get the index of the predictions and attach corresponding dates
    pred_idx = np.arange(len(train_data_H) + self.lbck, len(train_data_H)+ self.lbck + len(self.y_test))
    df_predict = pd.DataFrame(self.y_pred_inv, columns=['predicted '+ lb for lb in self.list_labels])
    df_predict = pd.concat([df_predict, pd.DataFrame(pred_idx, columns=['index'])], axis=1)
    df_predict = df_predict.set_index('index')
    df_predict = pd.concat([DATE, df_predict], axis=1, join='inner')
    
    if combined: # all lines on the same plot
      if '*rGAS' in self.list_labels and len(self.list_labels) > 1: # gas requires second y axis
        fig = make_subplots(specs=[[{"secondary_y": True}]])        # oil and water rate can share one y axis
        for i, label in enumerate(self.list_labels):
          if label =='*rGAS': # setup secondary axis for gas
            fig.add_trace(go.Scattergl(x=df_train['*DATE'], y=df_train[label], name='Train ' + self.rate[i], mode='markers'), secondary_y=True)
            fig.add_trace(go.Scattergl(x=df_actual['*DATE'], y=df_actual[label], name='Actual test ' + self.rate[i], mode='markers'), secondary_y=True)
            fig.add_trace(go.Scattergl(x=df_predict['*DATE'], y=df_predict['predicted ' + label], name='Prediction ' + self.rate[i], mode='markers'), secondary_y=True)
            continue
          # water and oil lines
          fig.add_trace(go.Scattergl(x=df_train['*DATE'], y=df_train[label], name='Train ' + self.rate[i], mode='markers'))
          fig.add_trace(go.Scattergl(x=df_actual['*DATE'], y=df_actual[label], name='Actual test ' + self.rate[i], mode='markers'))
          fig.add_trace(go.Scattergl(x=df_predict['*DATE'], y=df_predict['predicted ' + label], name='Prediction ' + self.rate[i], mode='markers'))
          fig.update_layout(autosize=True,
                            height=800,
                            title="LSTM forecast and actual ",
                            xaxis_title="Date",
                            yaxis_title="Rate (m\u00b3)",
                            yaxis2_title='Rate (m\u00b3)',
                            font=dict(family="Courier New, monospace",
                                      size=26,
                                      color="#7f7f7f"))
        fig.show()
      else: # if gas by itself or no gas
        fig = go.Figure()
        for i, label in enumerate(self.list_labels):
          fig.add_trace(go.Scattergl(x=df_train['*DATE'], y=df_train[label], name='Train ' + self.rate[i], mode='markers'))
          fig.add_trace(go.Scattergl(x=df_actual['*DATE'], y=df_actual[label], name='Actual test ' + self.rate[i], mode='markers'))
          fig.add_trace(go.Scattergl(x=df_predict['*DATE'], y=df_predict['predicted ' + label], name='Prediction ' + self.rate[i], mode='markers'))
          fig.update_layout(autosize=True,
                            height=800,
                            title="LSTM forecast and actual",
                            xaxis_title="Date",
                            yaxis_title="Rate (m\u00b3)",
                            font=dict(family="Courier New, monospace",
                                      size=26,
                                      color="#7f7f7f"))
        fig.show()
    else: # sub plots
      if len(self.rate) == 3:
        fig = make_subplots(rows=len(self.rate), cols=1, subplot_titles=(self.rate[0], self.rate[1], self.rate[2]))
        fig.update_layout(height=1500, width=1800)
      elif len(self.rate) == 2:
        fig = make_subplots(rows=len(self.rate), cols=1, subplot_titles=(self.rate[0], self.rate[1]))      
        fig.update_layout(height=1200, width=1800)
      elif len(self.rate) == 1:
        raise Exception('Too few labels for subplot, use default combined=True.')

      for i, label in enumerate(self.list_labels):
        fig.add_trace(go.Scattergl(x=df_train['*DATE'], y=df_train[label], name='Train ' + self.rate[i], mode='markers', marker_color='blue'), row=i+1, col=1)
        fig.add_trace(go.Scattergl(x=df_actual['*DATE'], y=df_actual[label], name='Actual test ' + self.rate[i], mode='markers', marker_color='green'), row=i+1, col=1)
        fig.add_trace(go.Scattergl(x=df_predict['*DATE'], y=df_predict['predicted ' + label], name='Prediction ' + self.rate[i], mode='markers', marker_color='orange'), row=i+1, col=1)       
        fig.update_xaxes(title_text="Date", row=i+1, col=1)
        fig.update_yaxes(title_text='Rate (m\u00b3)', row=i+1, col=1)
        fig.update_layout(title="LSTM forecast and actual",
                          font=dict(family="Courier New, monospace",
                                    size=26,
                                    color="#7f7f7f"),
                          title_text="LSTM Predictions")
        
      for i in fig['layout']['annotations']:
            i['font'] = dict(size=26,color='#ff0000')
      fig.show()
    
  def log_mlflow(self, log_dir):
    '''
    Save model parameters and metrics in log_dir
    
    Parameter:
      log_dir: str
        location of mlflow experiment file
    '''
    
    mlflow.set_experiment(log_dir)
    mlflow.keras.autolog()
    mlflow.start_run()
    mlflow.log_param('run_name', self.mode + '_' + self.name)
    mlflow.log_param("learning_rate", self.lr)
    mlflow.log_param('units', self.units)
    mlflow.log_param('epochs', self.num_epoch)
    mlflow.log_param('lookback', self.lbck)
    mlflow.log_param('batch_size', self.batch_size)
    mlflow.log_param('dropout', self.drop)

    for i, label in enumerate(self.rate):
      mlflow.log_metric("test_rmse_"+self.rate[i], self.metrics[label]['rmse'])
      mlflow.log_metric("test_r2_"+self.rate[i], self.metrics[label]['r2'])
      mlflow.log_metric("test_mae_"+self.rate[i], self.metrics[label]['mae'])
    mlflow.end_run()    


# COMMAND ----------





'''Nicolas Trinephi - acse-nt719

Utility modules

This module contains necessary imports for the VFM_Main file. This file requires all libraries from
requirements.txt to be installed on the cluster.

Contains utils class and InputError class
'''

import numpy as np
import copy
import datetime
import math

# data loading
from pyspark.sql.types import *
from pyspark.sql.functions import col
import pandas as pd

# interactive plotting
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import plotly.offline as pyo
import panel as pn
import hvplot
import hvplot.pandas
import holoviews as hv
import panel as pn

class utils:
  '''
  Class with utility functions that are all public.
  
  Methods:
    list_flat() lambda function
      flattens a list
    list_deep() lambda fuinction
      nests elements of a list
    to_pandas(df_list)
      Converts PySpark DataFrame to Pandas DataFrame
    re_writer(file)
      Rewrites data file removing and editing unusable text
    dynamic_filter(df, bin_size, col_name, r_lim)
      Smoothens the curve of the data within df[col_name]
    find_well(well_name, day, month, test)
      Find dataframes of well named well_name
    keep_param(day, month, test)
      Keep useful paramters for the LSTM model
    un_accum(month)
      calculates and adds columns for the unaccumulated rates in monthly historic data
    nan_exc(df_day, df_month, df_test)
      Clears data that are out of the limits set after quality control
    rmv_exc(target, *args)
      Removes 0 values of the rates
    hf_prep_input(high_freq)
      Merges high frequency datasets together and removes none useful columns
    prep_input(target, lets_keep)
      creates DATE dataframe with time stamp of kept values and removes unwanted columns from target
  '''

  list_flat = lambda l: [item for sublist in l for item in sublist]
  list_flat.__doc__ = '''lamda function to flatten a nested list'''

  list_deep = lambda lst: [[i] for i in lst]
  list_deep.__doc__ = '''lambda function to nest a list'''

  def to_pandas(df_list):
    '''
    Converts PySpark DataFrame to Pandas DataFrame
    
    Parameters: 
      df_list: list of PySpark DataFrames
    '''

    df = []
    for i in range(len(df_list)):
      tmp = df_list[i].toPandas()
      tmp['*DATE'] =  pd.to_datetime(tmp['*DATE'], infer_datetime_format=True)
      df.append(tmp)
    return df

  def re_writer(file): # pythonexamples.org
    '''
    Rewrites data file removing and editing unusable text.
    
    Parameters:
      file: string
        text file to edit
    '''
    if 'FileStore' in  file:
      f = '/dbfs/' + file # to access DataBricks file system
    else:
      pass
    fin = open(f, "rt", encoding='cp1252') # this encoding can open special characters
    data = fin.read()
    data = data.replace(' -99999', '0')
    data = data.replace('null', '0')
    data = data.replace('Null', '0')
    data = data.replace('NO 6', 'NO_6')
    data = data.replace(' A', '_A')
    data = data.replace(' C', '_C')
    data = data.replace(' B', '_B')
    data = data.replace(' H', '_H')
    data = data.replace('OIL,CHOKE_SIZE_GI', 'rOIL *CHOKE_SIZE_GI')
    data = data.replace('GAS_RATE,PRODUCTIVITY_INDEX', 'rGAS *PRODUCTIVITY_INDEX')
    data = data.replace('Date', 'DATE')

    fin.close()
    #open the input file in write mode
    fin = open(f, "wt")
    #overrite the input file with the resulting data
    fin.write(data)
    fin.close()
    
  def dynamic_filter(df, bin_size, col_name, r_lim): # try different spread for day and month
    '''
    Smoothens the curve of the data within df[col_name]. 
    
    Smaller bin_size and r_lim will result in tighter, smoother curve.
    
    Parameters:
      df: pd.DataFrame
      bin_size: int
        size of bin where the current value will be compared to values inside the bin
      col_name: str
        name of column where studied data is stored
      r_lim: int
        limit which will determine if the current value is to be adjusted
    Outputs:
      dfilter: pd.DataFrame
        dataframe containing adjusted values
    '''
    dfilter = df.copy()
    for row in range(1, len(df)-1):
      cur = df.loc[row, col_name] # current value
      if cur == np.nan:
        continue

      # calculate past and futur means
      if row < bin_size:
        past = df.iloc[0:row,:].loc[:, col_name]
        fut = df.iloc[row+1:row+bin_size,:].loc[:, col_name]
        m_past = past.mean()
        m_fut = fut.mean()
      elif row > len(df)-bin_size:
        past = df.iloc[row-bin_size:row,:].loc[:, col_name]
        fut = df.iloc[row+1:len(df),:].loc[:, col_name]
        m_past = past.mean()
        m_fut = fut.mean()
      else:
        past = df.iloc[row-bin_size:row,:].loc[:, col_name]
        fut = df.iloc[row+1:row+bin_size,:].loc[:, col_name]
        m_past = past.mean()
        m_fut = fut.mean()

      # filter if nan
      if m_past == np.nan: # can change the default value
        if abs(cur-m_fut) > r_lim:
          dfilter.loc[row, col_name] = m_fut
          continue
        else:
          continue

      if m_fut == np.nan:
        if abs(cur-m_past) > r_lim:
          dfilter.loc[row, col_name] = m_past
          continue
        else:
          continue

      # now compare means to current value
      if abs(cur-m_past) < r_lim or abs(cur-m_fut) < r_lim:
        continue
      else:
        # look at m_diff
        m_array = np.array([m_past, m_fut])
        tmp = np.isnan(m_array) # np.isnan because bin may contain nan
        m_array[tmp] = 0.0
        m_diff = abs(np.diff(m_array))[0]
        if m_diff < r_lim:
          b_mean = np.nanmean(m_array)
          if abs(cur-b_mean) > r_lim:
            dfilter.loc[row, col_name] = b_mean
            continue
          else:
            continue
        else: # check choppy bins
          max_past = past.max()
          min_past = past.min()
          max_fut = fut.max()
          min_fut = fut.min()
          if max_fut - min_fut > r_lim*2:
            dfilter.loc[row, col_name] = m_past
            continue
          if max_past - min_past > r_lim*2:
            dfilter.loc[row, col_name] = m_fut
            continue
    return dfilter
    
  def find_well(well_name, day, month, test):
    '''
    Find dataframes of well named well_name
    
    Parameters:
      well_name: str            
        name of well of interest            
      day: data object  
      month: data object  
      test: data object   
    Outputs:   
      l_day: list of day dataframes  
      l_month: list of month dataframes  
      l_test: list of test dataframes  
      well_ids: list of well tail names  
      test_ids: list of well head names  
    '''
    l_day = []
    l_month = []
    l_test = []
    assert (month.list_ids == day.list_ids)
    print('month and day ids are identical.')
    
    # monthly and daily have same well ids
    index = [index for index, name in enumerate(day.list_ids) if well_name in name]
    well_ids = [day.list_ids[i] for i in index]
    l_day.append([day.list_df[i] for i in index])
    l_month.append([month.list_df[i] for i in index])
    
    # test well ids
    index = [index for index, name in enumerate(test.list_ids) if well_name in name]
    test_ids = [test.list_ids[i] for i in index]
    l_test.append([test.list_df[i] for i in index])

    l_day = utils.list_flat(l_day)
    l_month = utils.list_flat(l_month)
    l_test = utils.list_flat(l_test)
    return l_day, l_month, l_test, well_ids, test_ids
  
  def keep_param(day, month, test):
    '''
    Keep useful paramters for the LSTM model
    
    Parameters:
      day: list pd.DataFrame
         day data
      month: list pd.DataFrame
         month data
      test: list pd.DataFrame
        test data
      '''
    all_df = [day, month, test]
    for df in all_df:
      for i in range(len(df)):
        df[i] = df[i][['*DATE', '*BHP_AVG', '*BHT_AVG', '*WHP_AVG_P', '*WHT_AVG_P', '*rOIL', '*rGAS', '*WATER']].reset_index(drop=True)
        
  def un_accum(month):
    '''
    calculates and adds columns for the unaccumulated rates in monthly historic data.
    
    Parameters:
      month: pd.DataFrame
        Dataframe of monthly historic data
    Outputs:
      month: pd.DataFrame
        Dataframe of monthly historic data with unaccumulated rates
    '''
    month = month.sort_values('*DATE').reset_index()

    month['days_in_month'] = month['*DATE'].dt.daysinmonth
    month = month.rename(columns={'*rOIL': '*OIL_ACCUM', '*rGAS': '*GAS_ACCUM',
                                                    '*WATER': '*WAT_ACCUM'})
    month['*rOIL'] = month['*OIL_ACCUM']/month['days_in_month']
    month['*rGAS'] = month['*GAS_ACCUM']/month['days_in_month']
    month['*WATER'] = month['*WAT_ACCUM']/month['days_in_month']
    return month
  
  def nan_exc(df_day, df_month, df_test):
    '''
    Clears data that are out of the limits set after quality control
    
    Parameters:
      df_day: pd.DataFrame
        dataframes containing day data
      df_month: pd.DataFrame
        dataframes containing month data
      df_test: pd.DataFrame
        dataframes containing test data
    '''
    df_list = [df_day, df_month, df_test]
  # remove exceptional data
    for df in df_list:
      df.loc[df['*WHP_AVG_P'] < 50, '*WHP_AVG_P'] = np.nan
      df.loc[df['*WHT_AVG_P'] < 20, '*WHT_AVG_P'] = np.nan
      df.loc[df['*BHP_AVG'] < 20, '*BHP_AVG'] = np.nan
      df.loc[df['*BHT_AVG'] < 20, '*BHT_AVG'] = np.nan
      df.loc[df['*WHP_AVG_P'] > 400, '*WHP_AVG_P'] = np.nan
      df.loc[df['*WHT_AVG_P'] > 200, '*WHT_AVG_P'] = np.nan
      df.loc[df['*BHP_AVG'] > 400, '*BHP_AVG'] = np.nan
      df.loc[df['*BHT_AVG'] > 200, '*BHT_AVG'] = np.nan
        
  def rmv_exc(target, *args):
    '''
    Removes 0 values of the rates
    
    Shut in values are not interesting.
    
    target: list pd.DataFrame
      list of dataframes containing data of the well of interest
    *args: list
      strings of parameter to remove from target dataframe
        expected: 'remove_oil', 'remove_gas', 'remove_water', '*BHP_AVG',
                  '*BHT_AVG', '*WHP_AVG_P', '*WHT_AVG_P'
    '''
    
    for i in range(len(target)):
      if 'remove_oil' in args:
        target[i] = target[i][(target[i]['*rOIL'] > 0)] 
      elif 'remove_gas' in args:
        target[i] = target[i][target[i]['*rGAS'] > 0]
      elif 'remove_water' in args:
        target[i] = target[i][target[i]['*WATER'] > 0]
        
      # remove nans if obj are processed
      elif '*BHP_AVG' in args:
        target[i].dropna(subset=['*BHP_AVG'], inplace=True)
      elif '*BHT_AVG' in args:
        target[i].dropna(subset=['*BHT_AVG'], inplace=True)
      elif '*WHP_AVG_P' in args:
        target[i].dropna(subset=['*WHP_AVG_P'], inplace=True)
      elif '*WHT_AVG_P' in args:
        target[i].dropna(subset=['*WHT_AVG_P'], inplace=True)
      elif args[i] not in ['remove_oil', 'remove_gas', 'remove_water', '*BHP_AVG',\
                  '*BHT_AVG', '*WHP_AVG_P', '*WHT_AVG_P']:
        try:
          raise InputError(args[i])
        except InputError as e:
          print('InputError: ', e.val, 'is not an expected input.')
    return target

  def hf_prep_input(high_freq):
    '''
    Merges high frequency datasets together and removes none useful columns
  
    Saves DATE dataframe with dates of each kept value
    
    Parameters:
      high_freq: list of pd.DataFrames
        high frequency dataframes 
    Output:
      hf_data: pd.DataFrame
        high frequency data ready for to be input into keras model
      DATE: pd.DataFrame
        date indexes
    '''
    # merge high frequency datasets
    hf_data = pd.merge_ordered(high_freq[0], high_freq[1], on=['*DATE'], how='outer').interpolate()
    hf_data = pd.merge_ordered(hf_data, high_freq[2], on=['*DATE'], how='outer').interpolate()
    hf_data = pd.merge_ordered(hf_data, high_freq[3], on=['*DATE'], how='outer').interpolate()

    # we want to use the data while there is no shut in, before July 26, 2013
    hf_data = hf_data[(hf_data['*DATE'] < datetime.datetime(2013,7,26))]
    hf_data = hf_data.fillna(0)

    DATE = pd.DataFrame(hf_data['*DATE']).reset_index(drop=True)

    hf_data.drop(hf_data.columns[[0,2,4,6]], axis=1, inplace=True) # remove date and non usable columns
    hf_data = hf_data.reset_index(drop=True)
    return hf_data, DATE
  
  def prep_input(target, lets_keep):
    '''
    creates DATE dataframe with time stamp of kept values and removes unwanted columns from target
    
    Parameters:
      target: pd.DataFrame
      lets_keep: list of str
        column names of cols that will be kept
    Outputs:
      target: list pd.DataFrame
        Dataframe with usable columns
      DATE: list pd.DataFrame
        date indexes
    '''
    DATE = [pd.DataFrame(target[i]['*DATE']).reset_index(drop=True) for i in range(len(target))]
    target = [target[i][lets_keep].reset_index(drop=True) for i in range(len(target))] # keep useful columns
    return target, DATE
  
class InputError(Exception):
  '''
  Error class inheriting form Exception class
  
  Attributes:
    val
      Variable causing exception
  '''
  def __init__(self, val):
    '''
    Paramaters:
      val:
        Variable causing exception
      '''

    self.val = val
  def __str__(self):
    '''
    Print val
    '''

    return repr(self.val)

# COMMAND ----------





'''Nicolas Trinephi - acse-nt719

Setup file for PyPI upload
'''

import setuptools
from distutils.core import setup


with open("README.md", "r") as f:
    long_description = f.read()

setup(
	name='vfm_tool',
	author='Nicolas Trinephi',
	author_email='nicolas.trinephi@imperial.ac.uk',
	version='1.13',
	description='Individual MSc Project',
	packages=['vfm_tool'],
	license='MIT',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url='https://github.com/acse-2019/irp-acse-nt719/tree/master/Code/',
	include_package_data=True
	)


# Nicolas Trinephi - acse-nt719
# MAGIC %run /Users/nicolas.trinephi@imperial.ac.uk/VFM_Code/pandas_data

'''Spark DataFrame creator and manipulator

This module allows the user to load data into spark DataFrames.
It requires pyspark, this is readily available on Databricks clusters, as well as the other
dependencies in requirements.txt.

This module accepts space separated value files (.dat).

Contains spark_data class
'''
from vfm_tool.Utils import *
from vfm_tool.Pandas_data import pandas_data

from pyspark.sql.types import *
from pyspark.sql.functions import col

# setup spark environment
from pyspark.sql import SparkSession

spark = SparkSession \
    .builder \
    .appName("VFM Session") \
    .config("spark.some.config.option", "some-value") \
    .getOrCreate()
# import org.apache.spark as spark

class spark_data(pandas_data):
  '''
  Inherited class of pandas_data for spark data frames
  
  
  Methods:
    df_lister(id_col)
      Creates list of DataFrames, one for each distinct entry in id_col
    make_df()
      Creates PySpark DataFrames from Month, Day and Test Data. 
    high_freq(file_list, well_bore)    Class method
      Creates DataFrames for high frequency pressure data, in spark_data object
  '''
  def df_lister(self, id_col):
    '''
    Creates list of DataFrames, one for each distinct entry in id_col
    Parameters:
      id_col: string
        column with entries to be listed
    '''

    dict_ids = [list(x.asDict().values()) for x in self.df.select(id_col).distinct().collect()] # dict obj
    self.list_ids = []
    for i in range(len(dict_ids)):
      self.list_ids.append(str(dict_ids[i][0])) # keep str values
    # create list of DataFrames
    self.list_df = [self.df.where(self.df[id_col] == x) for x in self.list_ids]

  def make_df(self):
    '''
    Creates PySpark DataFrames from Month, Day and Test Data. 
    
    Creates custom schema for data input, filters unwanted data and returns 
    a list of wells and corresponding DataFrames.
    
    Parameters: 
      file_location: string
    Outputs:
      spark_df.df_lister outputs
    '''

    # create custom schema for specific data sheet types
    if '_DAY.' in self.file_location: # old day data, should not be used much
      print('***********Accepted day data (pre-2012)***********')
      cSchema = StructType([StructField("*WELLID", StringType(), True), 
                               StructField("*DATE", IntegerType(), True),
                               StructField("*BHP", FloatType(), True),
                               StructField("*BHP_TVD", FloatType(), True),
                               StructField("*CHOKE", FloatType(), True),
                               StructField("*DHP", FloatType(), True),
                               StructField("*DHT", FloatType(), True),
                               StructField("*rGAS", FloatType(), True), # change to *GAS if rates are not daily
                               StructField("*GASLIFT", FloatType(), True),
                               StructField("*GI", FloatType(), True),
                               StructField("*rOIL", FloatType(), True), # change to *OIL if rates are not daily
                               StructField("*UCP", FloatType(), True), # P & T to keep
                               StructField("*UCT", FloatType(), True),
                               StructField("*WATER", FloatType(), True),
                               StructField("*WI", FloatType(), True),
                               StructField("*UPTIME", FloatType(), True)])
      # clear unusable values
      utils.re_writer(self.file_location, 'OIL', 'GAS')
      # read data
      self.df = spark.read.format("csv").option("header", "true").option("sep", " ").option("mode", "dropMalformed") \
            .schema(cSchema).load(file_location)
      self.df = self.df.filter(~col('*DATE').isin(['null']))

    elif '_DAY_' in self.file_location:
      print('***********Accepted day data***********')
      cSchema = StructType([StructField("*WELL_BORE_CODE", StringType(), True), 
                                 StructField("*DATE", DateType(), True),
                                 StructField("*ON_STREAM", FloatType(), True),
                                 StructField("*BHP_AVG", FloatType(), True),
                                 StructField("*BHT_AVG", FloatType(), True),
                                 StructField("*DP_TUBING", FloatType(), True),
                                 StructField("*ANNULUS_PRESS_AVG", FloatType(), True),
                                 StructField("*CHOKE_SIZE_AVG_P", FloatType(), True),
                                 StructField("*AVG_CHOKE_UOM", StringType(), True),
                                 StructField("*ALLOC_GASLIFT", FloatType(), True),
                                 StructField("*WHP_AVG_P", FloatType(), True),
                                 StructField("*WHT_AVG_P", FloatType(), True),
                                 StructField("*SAND_RATE", FloatType(), True), 
                                 StructField("*DP_CHOKE_SIZE", FloatType(), True),
                                 StructField("*rOIL", FloatType(), True),
                                 StructField("*COND", FloatType(), True),
                                 StructField("*rGAS", FloatType(), True),
                                 StructField("*WATER", FloatType(), True),
                                 StructField("*GI", FloatType(), True),
                                 StructField("*WI", FloatType(), True),
                                 StructField("*DHP", FloatType(), True),
                                 StructField("*DHT", FloatType(), True),
                                 StructField("*CHOKE_SIZE_AVG_GI", FloatType(), True),
                                 StructField("*CHOKE_SIZE_AVG_WI", FloatType(), True),
                                 StructField("*WHP_AVG_GI", FloatType(), True),
                                 StructField("*WHP_AVG_WI", FloatType(), True),
                                 StructField("*WHT_AVG_GI", FloatType(), True),
                                 StructField("*WHT_AVG_WI", FloatType(), True)])
      # clear unusable values
      utils.re_writer(self.file_location)
      # read data
      self.df = spark.read.format("csv").option("sep", " ").schema(cSchema).load(self.file_location)
      self.df = self.df.filter(~col('*DATE').isin(['null']))
      self.df = self.df.orderBy('*DATE')

    elif 'MONTH' in self.file_location: # month data
      print('***********Accepted month data***********')
      cSchema = StructType([StructField("*WELL_BORE_CODE", StringType(), True), 
                               StructField("*DATE", DateType(), True),
                               StructField("*rGAS", FloatType(), True), # change to *GAS if rates are not daily
                               StructField("*WATER", FloatType(), True),
                               StructField("*GI", FloatType(), True),
                               StructField("*WI", FloatType(), True),
                               StructField("*UPTIME", FloatType(), True),
                               StructField("*BHP_AVG", FloatType(), True), # P & T to keep
                               StructField("*BHT_AVG", FloatType(), True),
                               StructField("*DP_TUBING", FloatType(), True),
                               StructField("*ANNULUS_PRESS_AVG", FloatType(), True),
                               StructField("*ALLOC_GASLIFT", FloatType(), True),
                               StructField("*SAND_RATE", FloatType(), True),
                               StructField("*DP_CHOKE_SIZE", FloatType(), True),
                               StructField("*rOIL", FloatType(), True), # this is probably only oil production # change to *OIL,CHOKE_SIZE_GI if rates are not daily
                               StructField("*CHOKE_SIZE_GI", FloatType(), True),
                               StructField("*CHOKE_SIZE_WI", FloatType(), True),
                               StructField("*CHOKE_SIZE_P", FloatType(), True),
                               StructField("*WHP_GI", FloatType(), True),
                               StructField("*WHP_WI", FloatType(), True),
                               StructField("*WHP_AVG_P", FloatType(), True),
                               StructField("*WHT_GI", FloatType(), True),
                               StructField("*WHT_WI", FloatType(), True),
                               StructField("*WHT_AVG_P", FloatType(), True)])
      # clear unusable values
      utils.re_writer(self.file_location)
      # read data
      self.df = spark.read.format("csv").option("sep", " ").schema(cSchema).load(self.file_location)
      self.df = self.df.filter(~col('*DATE').isin(['null']))
      self.df = self.df.orderBy('*DATE')
    
    elif '_TEST_' in self.file_location: # well test data
      print('***********Accepted test data***********')
      cSchema = StructType([StructField("*WELL_CODE", StringType(), True), 
                               StructField("*DATE", DateType(), True),
                               StructField("*RESULT_NO", StringType(), True),
                               StructField("*END_DATE", StringType(), True), # some -99999
                               StructField("*CHOKE_SIZE", FloatType(), True),
                               StructField("*GOR", FloatType(), True),
                               StructField("*RESERVOIR_PRESS", FloatType(), True),
                               StructField("*rOIL", FloatType(), True),   # rates are daily
                               StructField("*COND_RATE", FloatType(), True),
                               StructField("*rGAS", FloatType(), True),
                               StructField("*PRODUCTIVITY_INDEX", FloatType(), True),
                               StructField("*WATERCUT", FloatType(), True),
                               StructField("*WHP_AVG_P", FloatType(), True),
                               StructField("*WHT_AVG_P", FloatType(), True),
                               StructField("*STRUCTURE", StringType(), True),
                               StructField("*FLUID_PI", FloatType(), True),
                               StructField("*GAS_LIFT", FloatType(), True),
                               StructField("*GAS_LIFT_GKGL", FloatType(), True),
                               StructField("*GAS_LIFT_RGL", FloatType(), True),
                               StructField("*HSV", FloatType(), True),
                               StructField("*QC_M", FloatType(), True),
                               StructField("*QG_M", FloatType(), True),
                               StructField("*QO_M", FloatType(), True),
                               StructField("*QW_M", FloatType(), True),
                               StructField("*SAND_TRAP", FloatType(), True),
                               StructField("*H2S", FloatType(), True),
                               StructField("*TDEV_PRESS", FloatType(), True),
                               StructField("*TDEV_TEMP", FloatType(), True),
                               StructField("*SAND_RATE", FloatType(), True),
                               StructField("*BHP_AVG", FloatType(), True),
                               StructField("*BHT_AVG", FloatType(), True),
                               StructField("*WATER", FloatType(), True)])
      # clear unusable values
      utils.re_writer(self.file_location)
      # read data
      self.df = spark.read.format("csv").option("sep", " ").schema(cSchema).load(self.file_location)
      self.df = self.df.filter(~col('*DATE').isin(['null', 'TEST_WELL', '*DATE']))
      self.df = self.df.orderBy('*DATE')
    else:
      try:
        raise InputError(self.location)
      except:
        print('InputError: Incorrect/No file chosen')
        
  @classmethod
  def high_freq(cls, file_list, well_bore):
    '''
    Creates DataFrames for high frequency pressure data, in spark_data object
    
    With current high frequency data, file_list[3] is the combined oil rate while file_list[:3] 
    contain the pressure values for each tail.
    
    Parameters:
      file_list: list of str
        list of high frequency data file locations
      well_bore: list of str
        well bore names of file contents
    '''
    cls = cls.__new__(cls)  # Does not call pandas_data.__init__()
    
    cls.list_df = []

    for i in range(3):
      cls.df = spark.read.format("csv").option("sep", ",").option("header", 'true')\
                    .schema(StructType([StructField("*DATE", StringType(), True),  # import as str
                                        StructField("*DHP "+well_bore[i], StringType(), True)]))\
                    .load(file_list[i])
      cls.list_df.append(cls.df)               
    cSchema = StructType([StructField("*DATE", StringType(), True), 
                          StructField("*rOIL", StringType(), True)])
    cls.df = spark.read.format("csv").option("sep", ",").option("header", 'true').schema(cSchema).load(file_list[3])
    cls.list_df.append(cls.df)

    cls.list_df = utils.to_pandas(cls.list_df)

    # trim excel data 
    for i in range(3):
      cls.list_df[i] = cls.list_df[i][['*DATE', '*DHP '+well_bore[i]]]
    cls.list_df[3] = cls.list_df[3][['*DATE', '*rOIL']]

    for i in range(3):
        cls.list_df[i]['*WELL_BORE_CODE'] = well_bore[i]

    # remove unit cells
    for i in range(4):
      cls.list_df[i].drop(cls.list_df[i].tail(1).index, inplace=True)
      cls.list_df[i].drop(cls.list_df[i].head(1).index, inplace=True)
      cls.list_df[i]['*DATE']= pd.to_datetime(cls.list_df[i]['*DATE']) # spark importing as string due to null
      cls.list_df[i] = cls.list_df[i].sort_values('*DATE', ascending=True)
    for i in range(3):
      cls.list_df[i]['*DHP '+well_bore[i]] = pd.to_numeric(cls.list_df[i]['*DHP '+well_bore[i]]) # spark importing as string due to 1st line of csv
    cls.list_df[3]['*rOIL'] = pd.to_numeric(cls.list_df[3]['*rOIL'])
    return cls

# COMMAND ----------


